/*
	Functions and classes for use with SWIG to generate scripting
	language wrapper for IRT Command Language (Icl).
	
	IRT Command Language (Icl)
	http://www.b-a-h.com/software/irt/icl/

	Author(s): Brad Hanson (http://www.b-a-h.com/)
	See the file LICENSE for information on usage and redistribution.

	Copyright (c) 2000-2002, Bradley A. Hanson
	
*/


#include "swig_icl.h"

#ifdef ETIRM_NO_DIR_PREFIX
#include "DiscreteNormalDist.h"
#include "Start3PL.h"
#include "ICCLogistic.h"
#include "ICRF_GPCM.h"
#include "ICRF_PCM.h"
#include "ItemDichotomous.h"
#include "etirm/ItemPolytomous.h"
#include "ItemParamPriorBeta4.h"
#include "ItemParamPriorLogNormal.h"
#include "ItemParamPriorNormal.h"
#else
#include "etirm/DiscreteNormalDist.h"
#include "etirm/Start3PL.h"
#include "etirm/ICCLogistic.h"
#include "etirm/ICRF_GPCM.h"
#include "etirm/ICRF_PCM.h"
#include "etirm/ItemDichotomous.h"
#include "etirm/ItemPolytomous.h"
#include "etirm/ItemParamPriorBeta4.h"
#include "etirm/ItemParamPriorLogNormal.h"
#include "etirm/ItemParamPriorNormal.h"
#endif

#include <string>

// Use normal random number generator from Boost library
// (www.boost.org).
#include <boost/random/normal_distribution.hpp>


namespace etirm {

// Prototype of local function
int ParamIndex(const char *paramname, const char *funcname);

/* Definitions of static data members */
template <class L>
Response Item<L>::notPresentedResponse = '.';
template <class L>
Response Item<L>::defaultFirstResponse = '0';

// Item types for 1, 2, and 3 parameter logistic models
typedef ItemDichotomous<DiscreteLatentDist<double>, ICCLogistic<3> > item3PL_type;
typedef ItemDichotomous<DiscreteLatentDist<double>, ICCLogistic<2> > item2PL_type;
typedef ItemDichotomous<DiscreteLatentDist<double>, ICCLogistic<1> > item1PL_type;

// Item types for polytomous items
typedef ItemPolytomous<DiscreteLatentDist<double>, ICRF_GPCM > itemGPCM_type;
typedef ItemPolytomous<DiscreteLatentDist<double>, ICRF_PCM > itemPCM_type;

// global variable holding information about a run defined in
// swig_etirm.cpp
extern SwigEtirmRun *gEtirmRun;

// Check that new_etirm has been called
void CheckRunInit()
{
	if (!gEtirmRun) throw RuntimeError("The new_items_dist command has not been executed", 0);

}


/*	
	Return a pointer to a item object.
	
	model		Model to use for item (ThreePL, TwoPL, OnePL, GPCM, PCM)
	index		Index of item in examinee response string (zero-offset)
	nrespcat	Number of response categories for polytomous items (ignored for dichtomous items)
	allPriors	Priors for a, b, and c parameters
	D			Logistic scaling constant (D=1.7 make logistic curve close to normal)
	latentDist	Latent variable distribution (discrete points used in SetLatentVarPoints call)
	funcname	Name of calling function (used in error messages)
*/
static item_type *CreateItem(IRTModel model, int index, int nrespcat, PriorVector &allPriors, 
	double D, DiscreteLatentDist<Real> &latentDist,
	const char *funcname)
{

	item_type *item;
	
	// Vector to hold default parameters assigned to item
	item_type::param_vector default_param(3);
	
	// Vector to hold prior assigned to item parameters
	PriorVector priors(3);
	switch (model)
	{
	case ThreePL:
	{
		ICCLogistic<3> icc(D);
		item = new item3PL_type(index, icc, &latentDist);
		
		// Set default parameters
		default_param[0] = 1.0; // default a parameter
		default_param[1] = 0.0; // default b parameter
		default_param[2] = 0.2; // default c parameter
		item->SetParameters(default_param);

		// Set default parameter priors
		for (int i=0; i<3; ++i)
		{
			if (allPriors[i])
			{
				priors[i] = CreatePrior(allPriors[i]->DistributionName(), 
					allPriors[i]->GetParameters(), funcname);
			}
			else
			{
				priors[i] = 0;
			}
		}
		item->SetPriors(priors);
	}
		break;
		
	case TwoPL:
	{
		ICCLogistic<2> icc(D, 0.0);
		item = new item2PL_type(index, icc, &latentDist);
		
		// Set default parameters
		default_param.newsize(2);
		default_param[0] = 1.0; // default a parameter
		default_param[1] = 0.0; // default b parameter
		item->SetParameters(default_param);

		// Set default priors for a and b parameters
		priors.resize(2);
		for (int i=0; i<2; ++i)
		{
			if (allPriors[i])
			{
				priors[i] = CreatePrior(allPriors[i]->DistributionName(), 
					allPriors[i]->GetParameters(), funcname);
			}
			else
			{
				priors[i] = 0;
			}
		}
		item->SetPriors(priors);
	}
		break;
		
	case OnePL:
	{
		ICCLogistic<1> icc(D, 1.0, 0.0);
		item = new item1PL_type(index, icc, &latentDist);
		
		// Set default parameter
		default_param.newsize(1);
		default_param[0] = 0.0; // default b parameter

		// Set default prior for b parameter
		priors.resize(1);
		if (allPriors[1])
		{
			priors[0] = CreatePrior(allPriors[1]->DistributionName(), 
				allPriors[1]->GetParameters(), funcname);
		}
		else
		{
			priors[0] = 0;
		}

		item->SetPriors(priors);
		item->SetParameters(default_param);
	}
		break;
		
	case GPCM:
	{
		ICRF_GPCM icrf(nrespcat, item_type::DefaultFirstResponse());
		item = new itemGPCM_type(index, icrf, &latentDist);
		
		default_param.newsize(nrespcat);
		priors.resize(nrespcat);
		
		// Set default parameter and prior for a
		default_param[0] = 1.0;
		if (allPriors[0])
		{
			priors[0] = CreatePrior(allPriors[0]->DistributionName(), 
					allPriors[0]->GetParameters(), funcname);
		}
		else priors[0] = 0;
		
		// Set default parameters and priors for b
		for (int i=1; i<nrespcat; ++i)
		{
			default_param[i] = 0.0;
			if (allPriors[1])
			{
				priors[i] = CreatePrior(allPriors[1]->DistributionName(), 
					allPriors[1]->GetParameters(), funcname);
			}
			else priors[i] = 0;
		}

		item->SetPriors(priors);
		item->SetParameters(default_param);
	}
		break;

	case PCM:
	{
		ICRF_PCM icrf(nrespcat, item_type::DefaultFirstResponse(), 1.0);
		item = new itemPCM_type(index, icrf, &latentDist);
		
		default_param.newsize(nrespcat-1);
		priors.resize(nrespcat-1);
		
		// Set default parameters and priors for b
		for (int i=0; i<nrespcat-1; ++i)
		{
			default_param[i] = 0.0;
			if (allPriors[1])
			{
				priors[i] = CreatePrior(allPriors[1]->DistributionName(), 
					allPriors[1]->GetParameters(), funcname);
			}
			else priors[i] = 0;
		}

		item->SetPriors(priors);
		item->SetParameters(default_param);
	}
		break;
	default:
		throw RuntimeError("Invalid model for item", funcname);
	}
	
	return item;

}

// Holds information needed in creating item objects.
// These default values are used when new_etirm is called.
// Also holds random number objects used in generating normal random variables
class DefaultValues
{

public:

	DefaultValues();
	~DefaultValues();

	double D;
		// Logistic function scaling constant. D=1.7 makes logistic similar to
		// normal.
	
	PriorVector priors;
		// Vector of pointers to priors for a, b, and c parameters.
		
	IRTModel dichModel;
		// Default model to use for dichotomous items (ThreePL, TwoPL, OnePL)
	
	IRTModel polyModel;
		// Default model to use for polytomous items (GPCM, PCM)
	
	int baseGroup;
		// Examinee group to use as base group for which
		// mean and s.d. of latent variable distribution are
		// set at zero and one.

	random_type *base_rand_normal;
		// Base random number generator object for normal random generator
		
	boost::normal_distribution<random_type> *rand_normal;
		// Normal random number generator object 

	
};

// global variable holding default values used in creating item objects
DefaultValues gIclDefaults;

// Set initial default values
DefaultValues::DefaultValues()
	: D(1.7), priors(3), dichModel(ThreePL), polyModel(GPCM), baseGroup(1), base_rand_normal(0), rand_normal(0)
{

	priors[0] = new ItemParamPriorBeta4(1.75, 3.0, 0.0, 3.0);
	priors[1]= new ItemParamPriorBeta4(1.01, 1.01, -6.0, 6.0);
	priors[2] = new ItemParamPriorBeta4(3.5, 4.0, 0.0, 0.5);

}

// Release memory allocated by constructor
DefaultValues::~DefaultValues()
{
	
	PriorVector::iterator ip = priors.begin();
	for (int n = priors.size(); n--; ++ip)
	{
		if (*ip) delete *ip;
	}
	
	if (base_rand_normal) delete base_rand_normal;
	if (rand_normal) delete rand_normal;
}

// Subclass of SwigEtirmRun to implement Icl
class SwigIclRun : public SwigEtirmRun
{

public:

	SwigIclRun(int nitems, int_vector *ncat, int nlatentcat, int ngroups, double minTheta, double maxTheta,
		bool uniquePoints);
	virtual ~SwigIclRun();

};

// SwigIclRun constructor
// nitems		Total number of items
// ncat			Vector giving number of response categories for each item (1 = dichotomous item,
//				>1 = polytomous item).
// nlatentcat	Number of categories of discrete latent variable
// ngroups		Number of examinee groups
// minTheta		Minimum point of discrete latent variable distribution
// maxTheta		Maximum point of discrete latent variable distribution
// uniquePoints If true use different latent distribution points for different examinee groups
SwigIclRun::SwigIclRun(int nitems, int_vector *ncat, int nlatentcat, int ngroups, double minTheta, 
	double maxTheta, bool uniquePoints)
		: SwigEtirmRun(nitems, nlatentcat, ngroups, minTheta, maxTheta, uniquePoints)
{

	int i;

	// Create items
	for (i=0; i<numItems; i++) 
	{
		IRTModel model = gIclDefaults.dichModel;
		int n = 1;
		if (ncat)
		{
			n = (*ncat)[i];
			if (n > 1) model = gIclDefaults.polyModel;
		}
		items[i] = CreateItem(model, i, n, gIclDefaults.priors , gIclDefaults.D, latentDist, 
			"SwigIclRun::SwigIclRun");

		/* Set up Uncmin objects to use for optimization in M-step */
		minProc[i] = new Uncmin<RealVector, RealMatrix, item_type>(items[i]->NumParameters());
		
		itemStats[i] = new ItemRespCounts(items[i]->NumRespCat(),ngroups);
	}

}

// Release memory allocated by constructor
SwigIclRun::~SwigIclRun()
{
}

/* translate string containing item parameter name to zero-offset index.
	Prototype is in swig_etirm.h */
int ParamIndex(const char *paramname, const char *funcname)
{
	switch (*paramname)
	{
		case 'a' : return 0;
		case 'b' : return 1;
		case 'c' : return 2;
	}
	
	char errstr[100];
	std::sprintf(errstr, "Invalid item parameter name: %.6s", paramname);
	throw RuntimeError(errstr, funcname);

	return -1;
}


/** Functions wrapped by SWIG **/

// Initialize memory to store information for items and the latent variable
// distributions. Create new SwigIclRun object.
// nitems		Total number of test items.
// ntheta		Number of discrete categories in latent variable distribution
// ngroup		Number of examinee groups. If > 1 than multiple group estimation is used
// ilist4		Vector containing number of response categories for each item (1 = dichotomous
//				item modeled by dichotomous model, >2 = polytomous item with that number of
//				response categories)
// dlist5		Vector containing minimum and maximum points to use for discrete 
//				latent variable distribution.
// uniquePoints If nonzero then unique latent distribution points are used for each examinee group
void new_items_dist(int nitems, int ntheta, int ngroups, int_vector *ilist4 , double_vector *dlist5, int uniquePoints)
{
	const char *funcname = "new_items_dist";	
	double minTheta,maxTheta; // theta minimum and maximum
	
	if (gEtirmRun) delete gEtirmRun;
	
	if (nitems < 0 || ntheta < 0 || ngroups < 0)
	{
		throw InvalidArgument("Invalid negative argument", funcname);
	}
	
	if (gIclDefaults.baseGroup > ngroups)
	{
		throw InvalidArgument("Number of groups specified makes default group invalid",
			funcname);
	}
	
	int_vector *models;
	if (ilist4 != 0)
	{
		if (ilist4->size() != nitems)
		{
			throw InvalidArgument("Number of element in response categories list does not equal number of items",
				funcname);
		}
		models = ilist4;
	}
	else
	{
		// If list of models is not specified use dichotomous model for all items
		models = new int_vector(nitems, 1);
	}
	
	if (dlist5 == 0)
	{
		minTheta = -4.0;
		maxTheta = 4.0;
	}
	else if (dlist5->size() != 2)
	{
		if (!ilist4) delete models;
		throw InvalidArgument("Fifth argument must be a list containing two elements",
			funcname);
	}
	else
	{
		minTheta = (*dlist5)[0];
		maxTheta = (*dlist5)[1];
	}
	
	// If there is only one group uniquePoints should be zero
	if (uniquePoints && ngroups == 1) uniquePoints = 0;

	gEtirmRun = new SwigIclRun(nitems, models, ntheta, ngroups, minTheta, maxTheta, uniquePoints);
	
	if (!ilist4) delete models;
}

// End run and delete gEtirmRun object
void delete_items_dist()
{
	if (gEtirmRun) delete gEtirmRun;
	gEtirmRun = 0;
}

// Return base examinee group
int get_base_group()
{
	return gIclDefaults.baseGroup;
}

// Change base examinee group
void set_base_group(int group)
{
	if (group < 0 || (gEtirmRun != 0 && gEtirmRun->numGroups < group))
	{
		throw InvalidArgument("Invalid group", "set_base_group");
	}
	
	gIclDefaults.baseGroup = group;
}

// Change value used for D in creating item objects
void set_default_D(double D)
{
	gIclDefaults.D = D;
}

// Change default item parameter priors used in creating item objects
//
// param		Parameter to set prior for (a, b, c)
// priortype	Prior distribution to use as default (beta, normal, lognormal, none)
// priorparam	Parameters of default prior distribution.
void set_default_prior(const char *param, const char *priortype, double_vector *priorparam)
{
	const char *funcname = "default_prior";
	int i = ParamIndex(param, funcname);
	
	if ((gIclDefaults.priors)[i]) delete (gIclDefaults.priors)[i];
	std::string pstr(priortype);
	(gIclDefaults.priors)[i] = CreatePrior(pstr, *priorparam, funcname);
}

// Return type of default prior for one item parameter (normal, lognormal, beta, none)
//	paramname	Name of parameter (a, b, or c)
const char *get_default_prior_type(const char *paramname)
{
	const char *fname = "get_default_prior_type";
	CheckRunInit();
	std::string &priorName = gEtirmRun->returnString;
	int index = ParamIndex(paramname, fname);
	
	PriorVector::iterator pv = gIclDefaults.priors.begin() + index;

	if (*pv != 0)
	{
		priorName = (*pv)->DistributionName();
	}
	else
	{
		priorName = "none";
	}
	
	return priorName.c_str();
	
	
}

// Return vector of parameters of default prior for one item parameter
//	paramname	Name of parameter (a, b, or c)
double_vector *get_default_prior_param(const char *paramname)
{
	const char *fname = "get_default_prior_param";
	CheckRunInit();
	int index = ParamIndex(paramname, fname);
	
	PriorVector::iterator pv = gIclDefaults.priors.begin() + index;
	
	double_vector *v;
	if (*pv != 0)
	{
		int n = (*pv)->NumParameters();
	
		v = new double_vector(n);
	
		*v = (*pv)->GetParameters();
	}
	else
	{
		v = new double_vector();
	}
	
	return v;
}


// Change default model used in creating item objects
//
// model		Model to set as default (ThreePL, TwoPL, OnePL)
void set_default_model_dichtomous(IRTModel irtModel)
{	
	gIclDefaults.dichModel = irtModel;
}

// Change default model used in creating item objects
//
// model		Model to set as default (ThreePL, TwoPL, OnePL)
void set_default_model_polytomous(IRTModel irtModel)
{	
	gIclDefaults.polyModel = irtModel;
}


// Set model to use for an item
//	itemno	Number of item to set model for (1-offset)
//	model	Model to use for item (3PL, 2PL, 1PL)
void item_set_model(int itemno, IRTModel irtModel)
{
	const char *fname = "item_set_model";
	CheckRunInit();
	gEtirmRun->CheckItemNumber(itemno, fname);

	item_type *item = (gEtirmRun->items)[itemno-1];
	
	int ncat = item->NumRespCat();
	
	// Don't let polytomous item be assigned a dichotomous model
	if (ncat > 2)
	{
		if (irtModel == ThreePL || irtModel == TwoPL || irtModel == OnePL)
		{
			throw InvalidArgument("Cannot assign a dichotomous model to a polytomous item", fname);
		}
	}
	
	// erase current information about item
	int index = item->Index();
	item->DeletePriors();
	delete item;
	
	// Recreate item information with new model
	item = CreateItem(irtModel, index, ncat, gIclDefaults.priors, gIclDefaults.D, 
		gEtirmRun->latentDist, fname);
	(gEtirmRun->items)[itemno-1] = item;
	
	// Recreate Uncmin object associated with item
	Uncmin<RealVector, RealMatrix, item_type> *uncmin = (gEtirmRun->minProc)[itemno-1];
	delete uncmin;
	uncmin = new Uncmin<RealVector, RealMatrix, item_type>(item->NumParameters());
	(gEtirmRun->minProc)[itemno-1] = uncmin;

}

/*
	item_3PL_starting_values
	
	Calculate item parameter starting values for 3PL, 2PL, and 1PL items.
	Returns number of items for which minimization procedure
	used to compute starting values failed.
	
	use_all
		If non-zero calculate PROX estimates for all examinees and items, even examinees
		who get all items right or all items wrong, and even for items answered all correctly 
		or all incorrectly. (default = 0)
		 
	item_all
		If non-zero all items are used to compute PROX estimates of examinee
		ability, otherwise only items for which starting values are computed
		are used. (default = 0)
	
	itemno3
		Vector of integers giving items numbers (1-offset)
		for which starting values are calculated. All these items must
		be modeled by the 3PL, 2PL, or 1Pl model. Note all 3PL items are
		still used to compute PROX estimates of examinee abilities which
		are used to compute the starting values for the items given by
		itemno3 if item_all is non-zero.
		If null compute starting values for all 3PL items (default = null).
		
*/
int item_3PL_starting_values(int use_all, int item_all, int_vector *itemno3)
{
	const char *fname = "item_starting_values";
	CheckRunInit();
	gEtirmRun->CheckExaminees(fname);
	
	int_vector *itemnos, *itemall3pl = 0;
	
	if (itemno3)
	{
		itemnos = itemno3;

		// Check that all items in list are 3PL, 2PL, or 1PL items
		ItemVector &all_items = gEtirmRun->items;
		int_vector::iterator ii = itemnos->begin();
		for (int i = itemnos->size(); i--; ++ii)
		{
			IRTModel m = all_items[*ii-1]->Model();
			if (m != ThreePL && m != TwoPL && m != OnePL)
			{
				throw InvalidArgument("Item number specified for which starting values can not be computed",
					"item_3PL_starting_values");
			}
		}
	}

	// Find item numbers for all 3PL, 2PL, and 1Pl items.
	if (item_all || !itemno3)
	{
		// Use std::vector to accumulate item numbers
		// since int_vector does not have a push_back function.
		std::vector<int> num;

		ItemVector::iterator ii = (gEtirmRun->items).begin();
		for (int i = (gEtirmRun->items).size(); i--; ++ii)
		{
			IRTModel m = (*ii)->Model();
			if (m == ThreePL || m == TwoPL || m == OnePL)
			{
				num.push_back((*ii)->Index()+1);
			}
		}
		
		itemall3pl = new int_vector(num.begin(), num.end());
		
		if (!itemno3) itemnos = itemall3pl;
	}
		

	UncminVector minsub = ItemSubset(itemnos,gEtirmRun->minProc, fname);
	ItemVector itemsub = ItemSubset(itemnos,gEtirmRun->items, fname);
	
	int n;
	try
	{
		if (item_all)
		{
			/* Create set containing pointers to items for which starting values are computed */
			std::set<item_type *> start_items;
			ItemVector::iterator ii = itemsub.begin();
			for (int i = itemsub.size(); i--; ++ii)
			{
				start_items.insert(*ii);
			}
	
			/* Vector of all 3PL items */
			ItemVector itemall = ItemSubset(itemall3pl, gEtirmRun->items, fname);

			n = StartingValues3PL<examinee_type::response_iterator, UncminVector::iterator,
				std::vector<examinee_type *>::iterator, item_type, std::vector<item_type *>::iterator>
				(minsub.begin(), (gEtirmRun->examinees).begin(), (gEtirmRun->examinees).end(), 
				itemall.begin(), itemall.end(), 
				item_type::NotPresentedResponse(), use_all,
				gIclDefaults.baseGroup, &start_items);
		}
		else
		{
			
			n = StartingValues3PL<examinee_type::response_iterator, UncminVector::iterator,
				std::vector<examinee_type *>::iterator, item_type, std::vector<item_type *>::iterator>
				(minsub.begin(), (gEtirmRun->examinees).begin(), (gEtirmRun->examinees).end(), 
				itemsub.begin(), itemsub.end(),  
				item_type::NotPresentedResponse(), use_all, gIclDefaults.baseGroup);
		}
	}
	catch (...)
	{
		if (itemall3pl) delete itemall3pl;
		throw;
	}

	
	if (itemall3pl) delete itemall3pl;

	return n;

}

// Set seed of random number generator used for simulating from a normal distribution
void normal_seed(unsigned long seed)
{
	
	// convert to type used for seed in boost random number library
	boost::uint32_t boost_seed = seed;
	
	if (gIclDefaults.base_rand_normal)
	{
		gIclDefaults.base_rand_normal->seed(boost_seed);
	}
	else
	{
		gIclDefaults.base_rand_normal = new random_type(boost_seed);
		gIclDefaults.rand_normal = new boost::normal_distribution<random_type>(*(gIclDefaults.base_rand_normal));
	}
}

// Return a generated number from a normal distribution with the specified mean and s.d.
double rand_normal(double mean, double sd)
{
	if (!gIclDefaults.base_rand_normal)
	{
		gIclDefaults.base_rand_normal = new random_type();
		gIclDefaults.rand_normal = new boost::normal_distribution<random_type>(*(gIclDefaults.base_rand_normal));
	}
	
	return (*gIclDefaults.rand_normal)() * sd + mean;

}


} // namespace etirm
